# Copyright (C) 2019-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import importlib
from typing import TYPE_CHECKING, Any, Dict
import warnings

if TYPE_CHECKING:
    # importing swh.storage.interface triggers the load of 300+ modules, so...
    from swh.search.interface import SearchInterface


SEARCH_IMPLEMENTATIONS = {
    "elasticsearch": ".elasticsearch.ElasticSearch",
    "remote": ".api.client.RemoteSearch",
    "memory": ".in_memory.InMemorySearch",
}


def get_search(cls: str, **kwargs: Dict[str, Any]) -> "SearchInterface":
    """Get an search object of class `cls` with arguments `args`.

    Args:
        cls: search's class, either 'local' or 'remote'
        args: dictionary of arguments passed to the
            search class constructor

    Returns:
        an instance of swh.search's classes (either local or remote)

    Raises:
        ValueError if passed an unknown search class.

    """
    if "args" in kwargs:
        warnings.warn(
            'Explicit "args" key is deprecated, use keys directly instead.',
            DeprecationWarning,
        )
        kwargs = kwargs["args"]

    class_path = SEARCH_IMPLEMENTATIONS.get(cls)
    if class_path is None:
        raise ValueError(
            "Unknown search class `%s`. Supported: %s"
            % (cls, ", ".join(SEARCH_IMPLEMENTATIONS))
        )

    (module_path, class_name) = class_path.rsplit(".", 1)
    module = importlib.import_module(module_path, package=__package__)
    Search = getattr(module, class_name)
    return Search(**kwargs)



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.search.utils import parse_and_format_date


@pytest.mark.parametrize(
    "date_str",
    ["2021-07-03", "2021-7-03", "2021-07-3", "2021-7-3", "2021-07-03T15:17:08Z"],
)
def test_parse_and_format_date_success(date_str):
    assert parse_and_format_date(date_str) == "2021-07-03"


@pytest.mark.parametrize(
    "date_str",
    ["foo", "2021/07/03", "2021+07+03T15,17,08Z"],
)
def test_parse_and_format_date_failure(date_str):
    assert parse_and_format_date(date_str) is None



# Copyright (C) 2019-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import socket
import subprocess
import time

import elasticsearch
import pytest

from swh.search import get_search

logger = logging.getLogger(__name__)


def free_port():
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("127.0.0.1", 0))
    port = sock.getsockname()[1]
    sock.close()
    return port


def wait_for_peer(addr, port):
    while True:
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.connect((addr, port))
        except ConnectionRefusedError:
            time.sleep(0.1)
        else:
            sock.close()
            break


CONFIG_TEMPLATE = """
node.name: node-1
path.data: {data}
path.logs: {logs}
network.host: 127.0.0.1
http.port: {http_port}
transport.port: {transport_port}
"""


def _run_elasticsearch(
    conf_dir, data_dir, logs_dir, http_port, transport_port, libffi_tmpdir
):
    es_home = "/usr/share/elasticsearch"

    with open(conf_dir + "/elasticsearch.yml", "w") as fd:
        fd.write(
            CONFIG_TEMPLATE.format(
                data=data_dir,
                logs=logs_dir,
                http_port=http_port,
                transport_port=transport_port,
            )
        )

    with open(conf_dir + "/log4j2.properties", "w") as fd:
        pass

    cmd = [
        "/usr/share/elasticsearch/jdk/bin/java",
        "-Des.path.home={}".format(es_home),
        "-Des.path.conf={}".format(conf_dir),
        "-Des.bundled_jdk=true",
        "-Dlog4j2.disable.jmx=true",
        # fix UnsupportedOperationException related to deprecated Security Manager
        # when using JDK >= 18
        "-Djava.security.manager=allow",
        "-cp",
        "{}/lib/*".format(es_home),
        "org.elasticsearch.bootstrap.Elasticsearch",
    ]

    host = "127.0.0.1:{}".format(http_port)

    with open(logs_dir + "/output.txt", "w") as fd:
        p = subprocess.Popen(cmd, env={"LIBFFI_TMPDIR": libffi_tmpdir})

    wait_for_peer("127.0.0.1", http_port)

    client = elasticsearch.Elasticsearch([host])
    assert client.ping()

    return p


@pytest.fixture(scope="session")
def elasticsearch_session(tmpdir_factory):
    tmpdir = tmpdir_factory.mktemp("elasticsearch")
    es_conf = tmpdir.mkdir("conf")

    http_port = free_port()
    transport_port = free_port()

    p = _run_elasticsearch(
        conf_dir=str(es_conf),
        data_dir=str(tmpdir.mkdir("data")),
        logs_dir=str(tmpdir.mkdir("logs")),
        http_port=http_port,
        transport_port=transport_port,
        libffi_tmpdir=str(tmpdir.mkdir("libffi")),
    )

    yield "127.0.0.1:{}".format(http_port)

    # Check ES didn't stop
    assert p.returncode is None, p.returncode

    p.kill()
    p.wait()


@pytest.fixture(scope="class")
def elasticsearch_host(elasticsearch_session):
    yield elasticsearch_session


@pytest.fixture
def swh_search(elasticsearch_host):
    """Instantiate a search client, initialize the elasticsearch instance,
    and returns it

    """
    logger.debug("swh_search: elasticsearch_host: %s", elasticsearch_host)
    search = get_search(
        "elasticsearch",
        hosts=[elasticsearch_host],
        indexes={
            "origin": {
                "index": "test",
                "read_alias": "test-read",
                "write_alias": "test-write",
            }
        },
    )
    search.deinitialize()  # To reset internal state from previous runs
    search.initialize()  # install required index
    yield search



# Copyright (C) 2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os
from typing import Any, Dict

import pytest
import yaml

from swh.core.api import RPCServerApp
from swh.core.config import load_from_envvar
from swh.search.api import server
from swh.search.api.server import load_and_check_config, make_app_from_configfile


def teardown_function():
    # Ensure there is no configuration loaded from a previous test
    server.api_cfg = None


def _write_config_file(tmp_path, monkeypatch, content):
    conf_path = os.path.join(str(tmp_path), "search.yml")
    with open(conf_path, "w") as f:
        f.write(yaml.dump(content))
    monkeypatch.setenv("SWH_CONFIG_FILENAME", conf_path)
    return conf_path


@pytest.fixture
def swh_search_server_config_without_indexes() -> Dict[str, Any]:
    return {
        "search": {
            "cls": "elasticsearch",
            "hosts": ["es1"],
        }
    }


@pytest.fixture
def swh_search_server_config_with_indexes(
    swh_search_server_config_without_indexes,
) -> Dict[str, Any]:
    return {
        "search": {
            **{"indexes": {"origin": {"index": "test"}}},
            **swh_search_server_config_without_indexes["search"],
        }
    }


@pytest.fixture
def swh_search_config_without_indexes(
    monkeypatch, swh_search_server_config_without_indexes, tmp_path
):
    return _write_config_file(
        tmp_path, monkeypatch, swh_search_server_config_without_indexes
    )


@pytest.fixture
def swh_search_config_with_indexes(
    monkeypatch, swh_search_server_config_with_indexes, tmp_path
):
    return _write_config_file(
        tmp_path, monkeypatch, swh_search_server_config_with_indexes
    )


def prepare_config_file(tmpdir, config_dict: Dict, name: str = "config.yml") -> str:
    """Prepare configuration file in `$tmpdir/name` with content `content`.

    Args:
        tmpdir (LocalPath): root directory
        content: Content of the file either as string or as a dict.
                            If a dict, converts the dict into a yaml string.
        name: configuration filename

    Returns
        path of the configuration file prepared.

    """
    config_path = tmpdir / name
    config_path.write_text(yaml.dump(config_dict), encoding="utf-8")
    # pytest on python3.5 does not support LocalPath manipulation, so
    # convert path to string
    return str(config_path)


@pytest.mark.parametrize("config_file", [None, ""])
def test_load_and_check_config_no_configuration(config_file):
    """Inexistent configuration files raises"""
    with pytest.raises(EnvironmentError, match="Configuration file must be defined"):
        load_and_check_config(config_file)


def test_load_and_check_config_inexistent_file():
    config_path = "/some/inexistent/config.yml"
    expected_error = f"Configuration file {config_path} does not exist"
    with pytest.raises(EnvironmentError, match=expected_error):
        load_and_check_config(config_path)


def test_load_and_check_config_wrong_configuration(tmpdir):
    """Wrong configuration raises"""
    config_path = prepare_config_file(tmpdir, {"something": "useless"})
    with pytest.raises(KeyError, match="Missing 'search' configuration"):
        load_and_check_config(config_path)


def test_load_and_check_config_local_config_fine(
    swh_search_server_config_with_indexes, tmpdir
):
    """'local' complete configuration is fine"""
    config_path = prepare_config_file(tmpdir, swh_search_server_config_with_indexes)
    cfg = load_and_check_config(config_path)
    assert cfg == swh_search_server_config_with_indexes


def test_server_make_app_from_config_file_without_indexes(
    swh_search_config_without_indexes,
):
    app = make_app_from_configfile()
    expected_cfg = load_from_envvar()

    assert app is not None
    assert isinstance(app, RPCServerApp)
    assert app.config["search"] == expected_cfg["search"]

    app2 = make_app_from_configfile()
    assert app is app2


def test_server_make_app_from_config_file_with_indexes(
    swh_search_config_with_indexes,
):
    app = make_app_from_configfile()
    expected_cfg = load_from_envvar()
    assert app is not None
    assert isinstance(app, RPCServerApp)
    assert app.config["search"] == expected_cfg["search"]

    app2 = make_app_from_configfile()
    assert app is app2


def test_server_first_call_initialize_elasticsearch(
    swh_search_config_with_indexes, mocker
):
    """Test the initialize method is called during the first and first only
    request to the server"""
    mock = mocker.patch("swh.search.elasticsearch.ElasticSearch.initialize")

    app = make_app_from_configfile()
    app.config["TESTING"] = True
    tc = app.test_client()

    tc.get("/")
    assert mock.call_count == 1

    tc.get("/")
    assert mock.call_count == 1



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

__all__ = ("SearchException", "SearchQuerySyntaxError")


class SearchException(Exception):
    """Base exception for errors specific to swh-search"""

    pass


class SearchQuerySyntaxError(SearchException):
    """Raised when the 'query' argument of origin_search cannot be parsed"""

    pass
